package de.tub.mapc.components.contestagents;

import java.util.Enumeration;
import java.util.List;

import org.slf4j.Logger;

import de.dailab.keks.io.IAddress;
import de.dailab.keks.io.IMessage;
import de.jiac.micro.agent.AbstractReactiveBehaviour;
import de.jiac.micro.agent.handle.ICommunicationHandle;
import de.jiac.micro.agent.memory.ILongTermMemory;
import de.jiac.micro.core.IAgent;
import de.jiac.micro.core.feature.ILifecycleAware;
import de.jiac.micro.core.scope.AgentScope;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.Coordination.goals.AbstractGoal;
import de.tub.mapc.Coordination.messages.SwitchAnfrage;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.components.contestagents.handle.IServerConnectorHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.knowledge.Switch;
import de.tub.mapc.monitor.AgentMonitor;
import de.tub.mapc.pathfinding.PathResult;

/**
 * reacts on perceptions concerning the grid
 * @author Erik
 *
 */
public class GridPerceptionElement extends AbstractReactiveBehaviour implements ILifecycleAware{

	
	private Logger _logger;
	private IAgent agent;

	private MessageService groupMessageService;
	private MessageService coordinatorMessageService;
	private IServerConnectorHandle _servercon;
	private ILongTermMemory memory;
	private IPathFindingHandle pathfinder;
	private Perception lastOwnPerception;
	
	@SuppressWarnings("unchecked")
	@Override
	protected Class[] filterDataTypes() {
		// we want to react on GridPerceptions (coming from the Contestserver)
		// and Messages (coming from other agents)
		return new Class[] {Perception.class, IMessage.class};
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void runShort(Enumeration sensorReadings) {
		while (sensorReadings.hasMoreElements()) {
			Object read = sensorReadings.nextElement();
			if (read instanceof Perception){
				// Perception of this Agent created from sensor Data 
				handlePerception((Perception)read);
			}else if (read instanceof IMessage) {
				// message created by other agent
				handleMessage((IMessage)read);
			}
		}
	}
	
	/**
	 * analyses the content of the received message and
	 * uses the information or reacts, depending on the content
	 * @param m the message
	 */
	private void handleMessage(IMessage m){
		Object content = m.getContent();
		if (content instanceof Perception){
			// perceptions of other agent -> update own worldmodel
			update((Perception)content);
		}else if (content instanceof AbstractGoal){
			AbstractGoal newGoal = (AbstractGoal) content;
			AbstractGoal lastGoal = (AbstractGoal) memory.rememberNewest(AbstractGoal.class);
			if (!newGoal.equals(lastGoal)){
				// --> received goal is really a new one
				
				// memorise it
				memory.memorise((AbstractGoal)content);
				
				// update monitor with goal
				try {
					AgentMonitor.getInstance().updateGoalList((AbstractGoal)content, agent.toString());
				}catch (NullPointerException e){
					_logger.error("GPE: ERROR when updating monitor with goal",e);
				}
			}else{
				// received the same goal again
				newGoal = lastGoal;
			}
			// find next action to achieve goal
			Gridworld world = (Gridworld) memory.rememberNewest(Gridworld.class);
			String action = newGoal.getAction(lastOwnPerception, world, pathfinder, coordinatorMessageService);
			
			// send action to contestserver
			sendActionToContestServer(action,lastOwnPerception.getID());
			
		}else if (content instanceof SwitchAnfrage){
			handleSwitchRequest((SwitchAnfrage)content);
		}
	}
	
	
	/**
	 * processes the perceptions of this agent, sent by the contestserver 
	 * @param p the perception
	 */
	private void handlePerception(Perception p){
		lastOwnPerception = p;

		// update own worldmodel
		
		// share the perception with other agents in the team
		// pruefen, ob das alte goal erreicht wurde und true/false mitsenden
		AbstractGoal lastGoal = (AbstractGoal) memory.rememberNewest(AbstractGoal.class);
		if (lastGoal != null){
			if(lastGoal.isAchieved())p.setReachedLastGoal(true);
		}else{
			p.setReachedLastGoal(true);
		}
		
		// perception teilen
		groupMessageService.sendPerception(p);
		
		// Update the monitor directly
//		AgentMonitor.getInstance().updateWorld(p, agent.toString());
		
	}

	
	/**
	 * Checks, wheter the Agent can reach a switching position of the requested switch without crossing an closed fence<br/>
	 * then sends an switchingAnswer to the coordinator, containing the path length to the switch
	 * or -1 if there is no switching position, that can be reached
	 * @param switchToBeOpened
	 */
	private void handleSwitchRequest(SwitchAnfrage switchToBeOpened){

		Gridworld world = (Gridworld) memory.rememberNewest(Gridworld.class);
		Position agentsPosition = new Position(lastOwnPerception.getPosX(), lastOwnPerception.getPosY());
		
		// get the position of the switch
		Switch s = world.getSwitches().get(switchToBeOpened.getId());
		List<Position> switchingPositions = s.getSwitchingPositions(world);
		int min = 999999;
		Position resultpos = null;
		for (Position openingPosition : switchingPositions) {
			PathResult pr = pathfinder.getPath(openingPosition, agentsPosition, world, false);
			if (pr.getErrortype() == null) {
				int length = pr.getPathLength();
				if (length < min) {
					resultpos = openingPosition;
					min = length;
				}
			}
		}
		if(resultpos == null){
			// there is no direct path, cannot open
//			_logger.info("Agent" + ID + " SWITCHREQUESTANSWER: -1");
			coordinatorMessageService.sendSwitchRequestAnswer(s.getX(), s.getY(), -1, null);
		}else{
//			_logger.info("Agent" + ID + " SWITCHREQUESTANSWER: "+min);
			coordinatorMessageService.sendSwitchRequestAnswer(s.getX(), s.getY(), min, resultpos);
		}
	}

	
	/**
	 * updates the own worldmodel with the GridFields seen by an agent<br/>
	 * and updates the pathfinder of this agent
	 * @param g the perception of an agent
	 * @return the updated gridworld
	 */
	private Gridworld update(Perception g){
		// access longTermMemory
		Gridworld world = (Gridworld) memory.rememberNewest(Gridworld.class);
		// and update it with the new GridInformation
		if(world != null)
			world.update(g.getGridInfo());
			
		return world;
	}
	
	/**
	 * sends an action to the contestserver using the serverconnector
	 * @param action a String describing the action, should not be null
	 * @param id the ID of the request-action message (perception)
	 */
	private void sendActionToContestServer(String action, String id){
		if(action == null){
			_logger.warn("calling serverconnector with action == null");
		}
		_servercon.act(action, null, id);
	}

	@Override
	public void start() {
        super.start();
        
		// Referenz auf den Agenten
		agent = AgentScope.getAgentReference();
		// Referenz auf den Logger
		_logger = agent.getLogger("GridPerceptor");
		// reference to communication module
		ICommunicationHandle _communicationHandle = (ICommunicationHandle) agent.getHandle(ICommunicationHandle.class);
		
		IAddress _coordinatorAddress= _communicationHandle.getMulticastAddressForName("coordinator");
		this.coordinatorMessageService = new MessageService(_communicationHandle, _coordinatorAddress);

		String groupName = (String) agent.getProperty(IMAPCConstants.TEAMNAME_KEY);
		IAddress _groupAddress = _communicationHandle.getMulticastAddressForName(groupName);
		this.groupMessageService = new MessageService(_communicationHandle, _groupAddress);		
		
		// reference to the Server connector
		_servercon = (IServerConnectorHandle) agent.getHandle(IServerConnectorHandle.class);
		// reference to the longtermmemory
		memory = (ILongTermMemory) agent.getHandle(ILongTermMemory.class);
		// reference to the pathfinder
		pathfinder = (IPathFindingHandle) agent.getHandle(IPathFindingHandle.class);
		if (pathfinder == null)_logger.error("PATHFINDER IST NULL");
	}

}
