package de.tub.mapc.components.contestagents;

import java.io.IOException;
import java.io.PushbackInputStream;

import org.slf4j.Logger;
import org.xmlpull.v1.XmlPullParserException;

import de.jiac.micro.agent.IActuator;
import de.jiac.micro.agent.ISensor;
import de.jiac.micro.agent.memory.IFact;
import de.jiac.micro.agent.memory.IShortTermMemory;
import de.jiac.micro.core.IAgent;
import de.jiac.micro.core.IHandle;
import de.jiac.micro.core.feature.AbstractPeriodicRunnable;
import de.jiac.micro.core.feature.ILifecycleAware;
import de.jiac.micro.core.handle.IResourceHandle;
import de.jiac.micro.core.io.IStreamConnection;
import de.jiac.micro.core.scope.AgentScope;
import de.jiac.micro.core.scope.ScopeFactory;
import de.jiac.micro.ext.io.ByteArrayOutputBuffer;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.components.contestagents.handle.IServerConnectorHandle;
import de.tub.mapc.parser.MAPCMessageFactory;
import de.tub.mapc.parser.MAPCMessageParser;

/**
 * This agent element provides the functionality to connect to the contest server.
 * 
 * @see IServerConnectorHandle
 * 
 * @author Tuguldur Erdene-Orchir
 * @author Marcel Patzlaff
 * @version $Revision: $
 */
public class ServerConnector implements IActuator, ISensor, ILifecycleAware, IServerConnectorHandle {
	/* The timer is used to reconnect if the connection between the contest server and the agent breaks. */
	protected final class ReconnectionTimer extends AbstractPeriodicRunnable {
        protected void runShort() {
            if(connect()) {
            	// disable() seemed not to function. -> Workaround with setPeriod()
            	//setPeriod(1000000000);            	
                disable();
            }
        }
    }
	
	private IAgent _agent;
	private Logger _logger;
	private IShortTermMemory _shortTermMemory;
	
	private ReconnectionTimer _reconnectionTimer;
	
	private String _serverIp;
	private String _serverPort;
	private String _teamName;
	private String _userName;
	private String _password;
	private int _agentID;
	
	private IStreamConnection _connection;
	
	private MAPCMessageParser _messageParser;
	private MAPCMessageFactory _messageFactory;
	
	private Runnable _streamReaderRunnable;
	private Thread _streamReaderThread;

	/* 12 kB buffer */
	private final byte[] _readBuffer= new byte[12 << 10];
	private PushbackInputStream _readStream;
    private final ByteArrayOutputBuffer _packetBuffer= new ByteArrayOutputBuffer();

	
	public IHandle getHandle() {		
		return this;
	}

	public void setShortTermMemory(IShortTermMemory shortTermMemory) {
		_shortTermMemory = shortTermMemory;		
	}

	public void initialise() { }

	public void start() {
		// Referenz auf den Agenten
		_agent = AgentScope.getAgentReference();
		// Referenz auf den Logger
		_logger = _agent.getLogger("ServerConnector");
		
		// Folgende Eigenschaften sind in der Agentendefinition definiert.
		_serverIp = (String) _agent.getProperty(IMAPCConstants.SERVER_IP_KEY);
		_serverPort = (String) _agent.getProperty(IMAPCConstants.SERVER_PORT_KEY);
		_teamName = (String) _agent.getProperty(IMAPCConstants.TEAMNAME_KEY);
		_userName = (String) _agent.getProperty(IMAPCConstants.USERNAME_KEY);
		_password = (String) _agent.getProperty(IMAPCConstants.PASSWORD_KEY);
		_agentID = Integer.valueOf((String) _agent.getProperty(IMAPCConstants.AGENT_ID_KEY)).intValue();
		
		_messageParser = new MAPCMessageParser(_logger, _userName, _agentID);
		_messageFactory = new MAPCMessageFactory();
		
		_streamReaderRunnable = new Runnable() {
			public void run() {
				try {
	            	while (_connection.isOpen()) {
	            		IFact fact = null;
	            		try {
	            			fact = _messageParser.parseXML(receivePacket());
	            		} catch (XmlPullParserException parserException) {
	            			_logger.warn("could not parse the message");
	            		}
	            		if (fact == null) {
	            			_logger.warn("something is wrong... Fact from received message is null");
// FIXME: das folgende funzt nicht. da wuerdet ihr ein paket an die debug meldung verlieren --marcel
//	            			_logger.debug(receivePacket());
	            		} else {
	            			// I am a sensor -> write into the memory that triggers the rules!
	            			_shortTermMemory.notice(fact);
	            		}
	    			}
	            } catch (IOException e) {
	                _logger.warn("connection error", e);
	                closeUnderlyingConnection();
	                _reconnectionTimer.setPeriod(1000000);
	                _reconnectionTimer.enable();
	            }
			}
		};
		
		/* The timer is used to reconnect if the connection between the contest server and the agent breaks. */
		_reconnectionTimer = new ReconnectionTimer();
		_reconnectionTimer.setPeriod(1000);
		_reconnectionTimer.enable();
	}

	public void stop() {
		_agent = null;
		_logger = null;
		
		_serverIp = null;
		_serverPort = null;
		_teamName = null;
		_userName = null;
		_password = null;
		
		_reconnectionTimer.disable();
		_reconnectionTimer = null;
		
		closeUnderlyingConnection();
		_streamReaderRunnable = null;
	}
	
	public void cleanup() { }
	
	private synchronized boolean connect() {
		if(_connection != null && _connection.isOpen()) {
            _logger.debug("connection is still alive");
        }
		_logger.debug("connecting to the server " + _serverIp + ":" + _serverPort + " with the username " + _userName + " and the password " + _password + " (the teamname " + _teamName +")...");
		String url= "socket://" + _serverIp + ":" + _serverPort;
		try {
			_connection = ((IResourceHandle) _agent.getHandle(IResourceHandle.class)).openStreamConnection(url);
			_readStream= new PushbackInputStream(_connection.getInputStream(), _readBuffer.length);
			_streamReaderThread = ScopeFactory.createScopeAwareThread(_streamReaderRunnable, "reader");
			_streamReaderThread.start();
			authenticate(_userName, _password);
			return true;
		} catch (IOException ioe) {
        	_logger.error("could not map socket connection", ioe);
            closeUnderlyingConnection();
        }
		return false;
	}
	
	private synchronized void closeUnderlyingConnection() {
		if(_connection != null) {
            _connection.close();
        }
        
        _connection = null;
        
        // avoid deadlock if this method was called out of the receiver thread
        if(_streamReaderThread != null && _streamReaderThread != Thread.currentThread()) {
            try {
            	_streamReaderThread.join();
            } catch (InterruptedException ie) {
                _logger.warn("could not join with reader thread");
            }
            
            _streamReaderThread = null;
        }
	}
	
	
	private byte[] receivePacket() throws IOException {
        if (!_connection.isOpen()) {
            _logger.error("connection already closed or not open");
            return new byte[0];
        }
        
        boolean nullFound= false;
        try {
	        while(!nullFound) {
	            int numBytes= _readStream.read(_readBuffer);
	            int nullByte= numBytes;
	            for(int n= 0; n < numBytes; ++n) {
	                if(_readBuffer[n] == 0) {
	                    nullFound= true;
	                    nullByte= n;
	                    break;
	                }
	            }
	            
	            _packetBuffer.write(_readBuffer, 0, nullByte);
	            
	            int unReadOff= nullByte + 1;
	            if(unReadOff < numBytes) {
	                _readStream.unread(_readBuffer, unReadOff, _readBuffer.length - unReadOff);
	            }
	        }

	        return _packetBuffer.toByteArray();
        } finally {
            _packetBuffer.reset();
        }
	}
	
	private void sendPacket(byte[] content) throws IOException {
		_connection.getOutputStream().write(content);
		_connection.getOutputStream().write(0);
		_connection.getOutputStream().flush();
	}
	
	public void authenticate(String username, String password) {
		try {
			sendPacket(_messageFactory.serialiseXML(_messageFactory.createAuthRequest(username, password)));
		} catch (IOException e) {
            _logger.warn("connection error", e);
            closeUnderlyingConnection();
            _reconnectionTimer.setPeriod(1000);
            _reconnectionTimer.enable();
        }
	}
	
	public void ping(String payload) {
		try {
			sendPacket(_messageFactory.serialiseXML(_messageFactory.createPing(payload)));
		} catch (IOException e) {
            _logger.warn("connection error", e);
            closeUnderlyingConnection();
            _reconnectionTimer.setPeriod(1000);
            _reconnectionTimer.enable();
        }
	}
	
	public void act(String type, String param, String id) {
		try {
			sendPacket(_messageFactory.serialiseXML(_messageFactory.createAction(type, param, id)));
		} catch (IOException e) {
            _logger.warn("connection error", e);
            closeUnderlyingConnection();
            _reconnectionTimer.setPeriod(1000);
            _reconnectionTimer.enable();
        }
	}

}
